from os.path import exists
from os.path import join as join_path
from typing import Any, Iterable, NamedTuple, TextIO
from warnings import warn

import numpy as np
from ase import Atoms
from ase.io import read, write
from ase.stress import voigt_6_to_full_3x3_stress
from pandas import DataFrame


def read_loss(filename: str) -> DataFrame:
    """Parses a file in `loss.out` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found `here
    <https://gpumd.org/nep/output_files/loss_out.html>`__.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) == 6:
        tags = 'total_loss L1 L2'
        tags += ' RMSE_P_train'
        tags += ' RMSE_P_test'
    elif len(data[0]) == 10:
        tags = 'total_loss L1 L2'
        tags += ' RMSE_E_train RMSE_F_train RMSE_V_train'
        tags += ' RMSE_E_test RMSE_F_test RMSE_V_test'
    elif len(data[0]) == 14:
        tags = 'total_loss L1 L2'
        tags += ' RMSE_E_train RMSE_F_train RMSE_V_train RMSE_Q_train RMSE_Z_train'
        tags += ' RMSE_E_test RMSE_F_test RMSE_V_test RMSE_Q_test RMSE_Z_test'
    else:
        raise ValueError(
            f'Input file contains {len(data[0])} data columns. Expected 6 or 10 columns.'
        )
    generations = range(100, len(data) * 100 + 1, 100)
    df = DataFrame(data=data[:, 1:], columns=tags.split(), index=generations)
    return df


def _write_structure_in_nep_format(structure: Atoms, f: TextIO) -> None:
    """Write structure block into a file-like object in format readable by nep executable.

    Parameters
    ----------
    structure
        input structure; must hold information regarding energy and forces
    f
        file-like object to which to write
    """

    # Allowed keyword=value pairs. Use ASEs extyz write functionality.:
    #   lattice="ax ay az bx by bz cx cy cz"                    (mandatory)
    #   energy=energy_value                                     (mandatory)
    #   virial="vxx vxy vxz vyx vyy vyz vzx vzy vzz"            (optional)
    #   weight=relative_weight                                  (optional)
    #   properties=property_name:data_type:number_of_columns
    #       species:S:1                      (mandatory)
    #       pos:R:3                          (mandatory)
    #       force:R:3 or forces:R:3          (mandatory)
    try:
        structure.get_potential_energy()
        structure.get_forces()  # calculate forces to have them on the Atoms object
    except RuntimeError:
        raise RuntimeError('Failed to retrieve energy and/or forces for structure')
    if np.isclose(structure.get_volume(), 0):
        raise ValueError('Structure cell must have a non-zero volume!')
    try:
        structure.get_stress()
    except RuntimeError:
        warn('Failed to retrieve stresses for structure')
    write(filename=f, images=structure, write_info=True, format='extxyz')


def write_structures(outfile: str, structures: list[Atoms]) -> None:
    """Writes structures for training/testing in format readable by nep executable.

    Parameters
    ----------
    outfile
        output filename
    structures
        list of structures with energy, forces, and (possibly) stresses
    """
    with open(outfile, 'w') as f:
        for structure in structures:
            _write_structure_in_nep_format(structure, f)


def write_nepfile(parameters: NamedTuple, dirname: str) -> None:
    """Writes parameters file for NEP construction.

    Parameters
    ----------
    parameters
        input parameters; see `here <https://gpumd.org/nep/input_parameters/index.html>`__
    dirname
        directory in which to place input file and links
    """
    with open(join_path(dirname, 'nep.in'), 'w') as f:
        for key, val in parameters.items():
            f.write(f'{key}  ')
            if isinstance(val, Iterable):
                f.write(' '.join([f'{v}' for v in val]))
            else:
                f.write(f'{val}')
            f.write('\n')


def read_nepfile(filename: str) -> dict[str, Any]:
    """Returns the content of a configuration file (`nep.in`) as a dictionary.

    Parameters
    ----------
    filename
        input file name
    """
    int_vals = ['version', 'neuron', 'generation', 'batch', 'population',
                'mode', 'model_type', 'charge_mode']
    float_vals = ['lambda_1', 'lambda_2', 'lambda_e', 'lambda_f', 'lambda_v',
                  'lambda_q', 'lambda_shear', 'force_delta']
    settings = {}
    with open(filename) as f:
        for line in f.readlines():
            # remove comments - throw away everything after a '#'
            cleaned = line.split('#', 1)[0].strip()
            flds = cleaned.split()
            if len(flds) == 0:
                continue
            settings[flds[0]] = ' '.join(flds[1:])
    for key, val in settings.items():
        if key in int_vals:
            settings[key] = int(val)
        elif key in float_vals:
            settings[key] = float(val)
        elif key in ['cutoff', 'n_max', 'l_max', 'basis_size', 'zbl', 'type_weight']:
            settings[key] = [float(v) for v in val.split()]
        elif key == 'type':
            types = val.split()
            types[0] = int(types[0])
            settings[key] = types
    return settings


def read_structures(dirname: str) -> tuple[list[Atoms], list[Atoms]]:
    """Parses the output files with training and test data from a nep run and returns their
    content as two lists of structures, representing training and test data, respectively.
    Target and predicted data are included in the :attr:`info` dict of the :class:`Atoms`
    objects.

    Parameters
    ----------
    dirname
        Directory from which to read output files.

    """
    path = join_path(dirname)
    if not exists(path):
        raise FileNotFoundError(f'Directory {path} does not exist')

    # fetch model type from nep input file
    nep_info = read_nepfile(f'{path}/nep.in')
    model_type = nep_info.get('model_type', 0)

    # set up which files to parse, what dimensions to expect etc
    # depending on the type of model that is parsed
    if model_type == 0:
        charge_mode = int(nep_info.get('charge_mode', 0))
        if charge_mode not in [0, 1, 2]:
            raise ValueError(f'Unknown charge_mode: {charge_mode}')
        # files to parse: (sname, size, mandatory, includes_target, per_atom)
        files_to_parse = [
            ('energy', 1, True, True, False),
            ('force', 3, True, True, True),
            ('virial', 6, True, True, False),
            ('stress', 6, True, True, False),
        ]
        if charge_mode in [1, 2]:
            # files to parse: (sname, size, includes_target, per_atom)
            files_to_parse += [
                ('charge', 1, True, False, True),
                ('bec', 9, False, True, True),
            ]
    elif model_type == 1:
        # files to parse: (sname, size, includes_target, per_atom)
        files_to_parse = [('dipole', 3, True, True, False)]
    elif model_type == 2:
        # files to parse: (sname, size, includes_target, per_atom)
        files_to_parse = [('polarizability', 6, True, True, False)]
    else:
        raise ValueError(f'Unknown model_type: {model_type}')

    # read training and test data
    structures = {}
    for stype in ['train', 'test']:
        filename = join_path(dirname, f'{stype}.xyz')
        try:
            structures[stype] = read(filename, format='extxyz', index=':')
        except FileNotFoundError:
            warn(f'File {filename} not found.')
            structures[stype] = []
            continue

        n_structures = len(structures[stype])

        # loop over files from which to read target data and predictions
        for sname, size, mandatory, includes_target, per_atom in files_to_parse:
            infile = f'{sname}_{stype}.out'
            path = join_path(dirname, infile)
            if not exists(path):
                if mandatory:
                    raise FileNotFoundError(f'File {path} does not exist')
                else:
                    continue
            ts, ps = _read_data_file(path, includes_target=includes_target)

            if ts is not None:
                if ts.shape[1] != size:
                    raise ValueError(f'Target data in {infile} has unexpected shape:'
                                     f' {ts.shape}  (expected: (-1, {size}))')
            if ps.shape[1] != size:
                raise ValueError(f'Predicted data in {infile} has unexpected shape:'
                                 f' {ps.shape}  (expected: (-1, {size}))')

            if per_atom:
                # data per-atom, e.g., forces, per-atom-virials, Born effective charges ...
                n_atoms_total = sum([len(s) for s in structures[stype]])
                if len(ps) != n_atoms_total:
                    raise ValueError(f'Number of atoms in {infile} ({len(ps)})'
                                     f' and {stype}.xyz ({n_atoms_total}) inconsistent.')
                n = 0
                for structure in structures[stype]:
                    nat = len(structure)
                    if ts is not None:
                        structure.info[f'{sname}_target'] = \
                            np.array(ts[n: n + nat]).reshape(nat, size)
                    structure.info[f'{sname}_predicted'] = \
                        np.array(ps[n: n + nat]).reshape(nat, size)
                    n += nat
            else:
                # data per structure, e.g., energy, virials, stress
                if len(ps) != n_structures:
                    raise ValueError(f'Number of structures in {infile} ({len(ps)})'
                                     f' and {stype}.xyz ({n_structures}) inconsistent.')
                for k, structure in enumerate(structures[stype]):
                    assert ts is not None, 'This should not occur. Please report.'
                    t = ts[k]
                    assert np.shape(t) == (size,)
                    structure.info[f'{sname}_target'] = t
                    p = ps[k]
                    assert np.shape(p) == (size,)
                    structure.info[f'{sname}_predicted'] = p

        # special handling of target data for BECs
        # The target data for BECs need not be complete. In this case nep writes
        # zeros for every component (not optimal). If we encounter such a case we set
        # all components to nan instead in order to be able to quickly filter for
        # this case when analyzing data.
        for s in structures[stype]:
            if 'bec_target' in s.info and np.allclose(s.info['bec_target'], 0):
                nat = len(s)
                size = 9
                s.info['bec_target'] = np.array(size * nat * [np.nan]).reshape(nat, size)

    return structures['train'], structures['test']


def _read_data_file(
    path: str,
    includes_target: bool = True,
):
    """Private function that parses *.out files and
    returns their content for further processing.
    """
    with open(path, 'r') as f:
        lines = f.readlines()
    target, predicted = [], []
    for line in lines:
        flds = line.split()
        if includes_target:
            if len(flds) % 2 != 0:
                raise ValueError(f'Incorrect number of columns in {path} ({len(flds)}).')
            n = len(flds) // 2
            predicted.append([float(s) for s in flds[:n]])
            target.append([float(s) for s in flds[n:]])
        else:
            predicted.append([float(s) for s in flds])
            target = None
    if target is not None:
        target = np.array(target)
    predicted = np.array(predicted)
    return target, predicted


def get_parity_data(
    structures: list[Atoms],
    property: str,
    selection: list[str] = None,
    flatten: bool = True,
) -> DataFrame:
    """Returns the predicted and target energies, forces, virials or stresses
    from a list of structures in a format suitable for generating parity plots.

    The structures should  have been read using :func:`read_structures
    <calorine.nep.read_structures>`, such that the `info` object is
    populated with keys of the form `<property>_<type>` where `<property>`
    is, e.g., `energy` or `force` and `<type>` is one of `predicted` or `target`.

    The resulting parity data is returned as a tuple of dicts, where each entry
    corresponds to a list.

    Parameters
    ----------
    structures
        List of structures as read with :func:`read_structures <calorine.nep.read_structures>`.
    property
        One of `energy`, `force`, `virial`, `stress`, `bec`, `dipole`, or `polarizability`.
    selection
        A list containing which components to return, and/or the norm.
        Possible values are `x`, `y`, `z`, `xx`, `yy`,
        `zz`, `yz`, `xz`, `xy`, `norm`, `pressure`.
    flatten
        if True return flattened lists; this is useful for flattening
        the components of force or virials into a simple list
    """
    voigt_mapping = {
        'x': 0, 'y': 1, 'z': 2, 'xx': 0, 'yy': 1, 'zz': 2,  'yz': 3, 'xz': 4, 'xy': 5,
    }
    if property not in ('energy', 'force', 'virial', 'stress', 'polarizability', 'dipole', 'bec'):
        raise ValueError(
            "`property` must be one of 'energy', 'force', 'virial', 'stress',"
            " 'polarizability', 'dipole', or 'bec'."
        )
    if property in ['energy'] and selection:
        raise ValueError('Selection cannot be applied to scalars.')
    if property != 'stress' and selection and 'pressure' in selection:
        raise ValueError(f'Cannot calculate pressure for `{property}`.')

    data = {'predicted': [], 'target': []}
    if property in ['force', 'bec'] and flatten:
        size = 3 if property == 'force' else 9
        data['species'] = []
    for structure in structures:
        if 'species' in data:
            data['species'].extend(np.repeat(structure.symbols, size).tolist())
        for stype in ['predicted', 'target']:
            property_with_stype = f'{property}_{stype}'
            if property_with_stype not in structure.info.keys():
                raise KeyError(f'{property_with_stype} not available in info field of structure')
            extracted_property = np.array(structure.info[property_with_stype])

            if selection is None or len(selection) == 0:
                data[stype].append(extracted_property)
                continue

            selected_values = []
            for select in selection:
                if property in ['force', 'bec']:
                    # flip to get (n_components, n_structures)
                    extracted_property = extracted_property.T
                if select == 'norm':
                    if property == 'force':
                        selected_values.append(np.linalg.norm(extracted_property, axis=0))
                    elif property in ['virial', 'stress']:
                        full_tensor = voigt_6_to_full_3x3_stress(extracted_property)
                        selected_values.append(np.linalg.norm(full_tensor))
                    elif property in ['dipole']:
                        selected_values.append(np.linalg.norm(extracted_property))
                    else:
                        raise ValueError(
                            f'Cannot handle selection=`norm` with property=`{property}`.')
                    continue

                if select == 'pressure' and property == 'stress':
                    total_stress = extracted_property
                    selected_values.append(-np.sum(total_stress[:3]) / 3)
                    continue

                if select not in voigt_mapping:
                    raise ValueError(f'Selection `{select}` is not allowed.')
                index = voigt_mapping[select]
                if index >= extracted_property.shape[0]:
                    raise ValueError(
                        f'Selection `{select}` is not compatible with property `{property}`.'
                    )
                selected_values.append(extracted_property[index])

            data[stype].append(selected_values)
    if flatten:
        for stype in ['target', 'predicted']:
            value = data[stype]
            if len(np.shape(value[0])) > 0:
                data[stype] = np.concatenate(value).ravel().tolist()
        if property in ['force']:
            n = len(data['target']) // 3
            data['component'] = ['x', 'y', 'z'] * n
        elif property in ['virial', 'stress']:
            n = len(data['target']) // 6
            data['component'] = ['xx', 'yy', 'zz', 'yz', 'xz', 'xy'] * n
        elif property in ['bec']:
            n = len(data['target']) // 9
            data['component'] = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz'] * n
    df = DataFrame(data)
    # In case of flatten, cast to float64 for compatibility
    # with e.g. seaborn.
    # Casting in this way breaks tensorial properties though,
    # so skip it there.
    if flatten:
        df['target'] = df.target.astype('float64')
        df['predicted'] = df.predicted.astype('float64')
    return df



from .analysis import (
    analyze_data,
    get_autocorrelation_function,
    get_correlation_length,
    get_error_estimate,
    get_rtc_from_hac,
)
from .entropy import get_entropy
from .phonons import get_force_constants
from .structures import (
    get_spacegroup,
    get_primitive_structure,
    get_wyckoff_sites,
    relax_structure,
)
from .stiffness import get_elastic_stiffness_tensor

__all__ = [
    'analyze_data',
    'get_autocorrelation_function',
    'get_correlation_length',
    'get_entropy',
    'get_error_estimate',
    'get_elastic_stiffness_tensor',
    'get_force_constants',
    'get_primitive_structure',
    'get_rtc_from_hac',
    'get_spacegroup',
    'get_wyckoff_sites',
    'relax_structure',
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_sitemap',
    'sphinx_design',
    'nbsphinx',
]

graphviz_output_format = 'svg'
templates_path = ['_templates']
default_role = 'code'
source_suffix = '.rst'
master_doc = 'index'
pygments_style = 'sphinx'
todo_include_todos = False

# Collect basic information from main module
metadata = importlib.metadata.metadata('calorine')
version = metadata.get('Version', 'n/a')
project = metadata.get('Name', 'n/a')
author = metadata.get('Author', 'n/a')

site_url = 'https://calorine.materialsmodeling.org/'
html_js_files = ['delete_empty_code_blocks.js',
                 'hidden_cells.js']
html_css_files = ['custom.css',
                  'hidden_cells.css']
html_logo = '_static/logo.png'
html_favicon = '_static/logo.ico'
html_static_path = ['_static']
htmlhelp_basename = 'calorinedoc'
intersphinx_mapping = {
    'ase':     ('https://wiki.fysik.dtu.dk/ase', None),
    'numpy':   ('https://numpy.org/doc/stable/', None),
    'h5py':    ('http://docs.h5py.org/en/latest/', None),
    'scipy':   ('https://scipy.github.io/devdocs/', None),
    'sklearn': ('https://scikit-learn.org/stable', None),
}

# sphinx theme configuration
html_theme = 'pydata_sphinx_theme'
html_theme_options = {
    'secondary_sidebar_items': ['page-toc'],
    'switcher': {
        'json_url': 'https://calorine.materialsmodeling.org/_static/switcher.json',
        'version_match': version,
    },
    'navbar_start': ['navbar-logo', 'version-switcher'],
}
html_sidebars = {
    'get_started': [],
    'credits': [],
    'genindex': [],
}

# Settings for nbsphinx
nbsphinx_execute = 'never'

# Options for LaTeX output
_PREAMBLE = r"""
\usepackage{amsmath,amssymb}
\renewcommand{\vec}[1]{\boldsymbol{#1}}
\DeclareMathOperator*{\argmin}{\arg\!\min}
\DeclareMathOperator{\argmin}{\arg\!\min}
"""

latex_elements = {
    'preamble': _PREAMBLE,
}
latex_documents = [
    (master_doc, 'calorine.tex', 'calorine Documentation',
     'The calorine developer team', 'manual'),
]


# Options for manual page output
man_pages = [
    (master_doc, 'calorine', 'calorine Documentation',
     [author], 1)
]


# Options for Texinfo output
texinfo_documents = [
    (master_doc, 'calorine', 'calorine Documentation',
     author, 'calorine', 'Strong coupling calculator',
     'Miscellaneous'),
]



from typing import List
import numpy as np
from ase import Atoms
from ase.filters import FrechetCellFilter
from ase.optimize import BFGS, LBFGS, FIRE, GPMin
from ase.optimize.sciopt import SciPyFminBFGS
from ase.units import GPa

try:
    import spglib
    spglib_available = True
except ImportError:  # pragma: no cover
    spglib_available = False


def relax_structure(structure: Atoms,
                    fmax: float = 0.001,
                    steps: int = 500,
                    minimizer: str = 'bfgs',
                    constant_cell: bool = False,
                    constant_volume: bool = False,
                    scalar_pressure: float = 0.0,
                    **kwargs) -> None:
    """Relaxes the given structure.

    Parameters
    ----------
    structure
        Atomic configuration to relax.
    fmax
        Stop relaxation if the absolute force for all atoms falls below this value.
    steps
        Maximum number of relaxation steps the minimizer is allowed to take.
    minimizer
        Minimizer to use; possible values: 'bfgs', 'lbfgs', 'fire', 'gpmin', 'bfgs-scipy'.
    constant_cell
        If True do not relax the cell or the volume.
    constant_volume
        If True relax the cell shape but keep the volume constant.
    kwargs
        Keyword arguments to be handed over to the minimizer; possible arguments can be found
        in the `ASE documentation <https://wiki.fysik.dtu.dk/ase/ase/optimize.html>`_
    scalar_pressure
        External pressure in GPa.
    """
    if structure.calc is None:
        raise ValueError('Structure has no attached calculator object')
    if constant_cell:
        ucf = structure
    else:
        ucf = FrechetCellFilter(
            structure, constant_volume=constant_volume, scalar_pressure=scalar_pressure * GPa)
    kwargs['logfile'] = kwargs.get('logfile', None)
    if minimizer == 'bfgs':
        dyn = BFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'lbfgs':
        dyn = LBFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'bfgs-scipy':
        dyn = SciPyFminBFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'fire':
        dyn = FIRE(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'gpmin':
        dyn = GPMin(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    else:
        raise ValueError(f'Unknown minimizer: {minimizer}')


def get_spacegroup(
    structure: Atoms,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
    style: str = 'international',
) -> str:
    """Returns the space group of a structure using spglib.
    This is a convenience interface to the :func:`get_spacegroup`
    function of spglib that works directly with ase Atoms objects.

    Parameters
    ----------
    structure
        Input atomic structure.
    symprec
        Tolerance imposed when analyzing the symmetry.
    angle_tolerance
        Tolerance imposed when analyzing angles.
    style
        Space group notation to be used. Can be ``'international'`` for the
        interational tables of crystallography (ITC) style (Hermann-Mauguin
        and ITC number) or ``'Schoenflies'`` for the Schoenflies notation.
    """
    if not spglib_available:
        raise ImportError('\
                spglib must be available in order for this function to work.')  # pragma: no cover

    if style == 'international':
        symbol_type = 0
    elif style == 'Schoenflies':
        symbol_type = 1
    else:
        raise ValueError(f'Unknown value for style: {style}')

    structure_tuple = (
        structure.get_cell(),
        structure.get_scaled_positions(),
        structure.numbers)
    spg = spglib.get_spacegroup(
        structure_tuple, symprec=symprec,
        angle_tolerance=angle_tolerance, symbol_type=symbol_type)

    return spg


def get_primitive_structure(
    structure: Atoms,
    no_idealize: bool = True,
    to_primitive: bool = True,
    symprec: float = 1e-5,
) -> Atoms:
    """Returns the primitive structure using spglib.
    This is a convenience interface to the :func:`standardize_cell`
    function of spglib that works directly with ase Atoms objects.

    Parameters
    ----------
    structure
        Input atomic structure.
    no_idealize
        If ``True`` lengths and angles are not idealized.
    to_primitive
        If ``True`` convert to primitive structure.
    symprec
        Tolerance imposed when analyzing the symmetry.
    """
    if not spglib_available:
        raise ImportError('\
                spglib must be available in order for this function to work.')  # pragma: no cover

    structure_tuple = (
        structure.get_cell(),
        structure.get_scaled_positions(),
        structure.numbers)
    result = spglib.standardize_cell(
        structure_tuple, to_primitive=to_primitive,
        no_idealize=no_idealize, symprec=symprec)
    if result is None:
        raise ValueError('spglib failed to find the primitive cell, maybe caused by large symprec.')
    lattice, scaled_positions, numbers = result
    scaled_positions = [np.round(pos, 12) for pos in scaled_positions]
    structure_prim = Atoms(scaled_positions=scaled_positions,
                           numbers=numbers, cell=lattice, pbc=True)
    structure_prim.wrap()

    return structure_prim


def get_wyckoff_sites(
        structure: Atoms,
        map_occupations: List[List[str]] = None,
        symprec: float = 1e-5,
        include_representative_atom_index: bool = False,
) -> List[str]:
    """Returns the Wyckoff symbols of the input structure.
    The Wyckoff labels can be conveniently attached as an array to the
    structure object as demonstrated in the examples section below.

    By default the occupation of the sites is part of the symmetry
    analysis. If a chemically disordered structure is provided this
    will usually reduce the symmetry substantially. If one is
    interested in the symmetry of the underlying structure one can
    control how occupations are handled. To this end, one can provide
    the :attr:`map_occupations` keyword argument. The latter must be a
    list, each entry of which is a list of species that should be
    treated as indistinguishable. As a shortcut, if *all* species
    should be treated as indistinguishable one can provide an empty
    list. Examples that illustrate the usage of the keyword are given
    below.

    Parameters
    ----------
    structure
        Input structure. Note that the occupation of the sites is
        included in the symmetry analysis.
    map_occupations
        Each sublist in this list specifies a group of chemical
        species that shall be treated as indistinguishable for the
        purpose of the symmetry analysis.
    symprec
        Tolerance imposed when analyzing the symmetry using spglib.
    include_representative_atom_index
        If True the index of the first atom in the structure that is
        representative of the Wyckoff site is included in the symbol.
        This is in particular useful in cases when there are multiple
        Wyckoff sites sites with the same Wyckoff letter.

    Examples
    --------
    Wyckoff sites of a hexagonal-close packed structure::

        >>> from ase.build import bulk
        >>> structure = bulk('Ti')
        >>> wyckoff_sites = get_wyckoff_sites(structure)
        >>> print(wyckoff_sites)
        ['2d', '2d']


    The Wyckoff labels can also be attached as an array to the
    structure, in which case the information is also included when
    storing the Atoms object::

        >>> from ase.io import write
        >>> structure.new_array('wyckoff_sites', wyckoff_sites, str)
        >>> write('structure.xyz', structure)

    The function can also be applied to supercells::

        >>> structure = bulk('GaAs', crystalstructure='zincblende', a=3.0).repeat(2)
        >>> wyckoff_sites = get_wyckoff_sites(structure)
        >>> print(wyckoff_sites)
        ['4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c',
         '4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c']

    Now assume that one is given a supercell of a (Ga,Al)As
    alloy. Applying the function directly yields much lower symmetry
    since the symmetry of the original structure is broken::

        >>> structure.set_chemical_symbols(
        ...        ['Ga', 'As', 'Al', 'As', 'Ga', 'As', 'Al', 'As',
        ...         'Ga', 'As', 'Ga', 'As', 'Al', 'As', 'Ga', 'As'])
        >>> print(get_wyckoff_sites(structure))
        ['8g', '8i', '4e', '8i', '8g', '8i', '2c', '8i',
         '2d', '8i', '8g', '8i', '4e', '8i', '8g', '8i']

    Since Ga and Al occupy the same sublattice, they should, however,
    be treated as indistinguishable for the purpose of the symmetry
    analysis, which can be achieved via the :attr:`map_occupations`
    keyword::

        >>> print(get_wyckoff_sites(structure, map_occupations=[['Ga', 'Al'], ['As']]))
        ['4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c',
         '4a', '4c', '4a', '4c', '4a', '4c', '4a', '4c']

    If occupations are to ignored entirely, one can simply provide an
    empty list. In the present case, this turns the zincblende lattice
    into a diamond lattice, on which case there is only one Wyckoff
    site::

        >>> print(get_wyckoff_sites(structure, map_occupations=[]))
        ['8a', '8a', '8a', '8a', '8a', '8a', '8a', '8a',
         '8a', '8a', '8a', '8a', '8a', '8a', '8a', '8a']
    """
    if not spglib_available:
        raise ImportError('\
                spglib must be available in order for this function to work.')  # pragma: no cover

    structure_copy = structure.copy()
    if map_occupations is not None:
        if len(map_occupations) > 0:
            new_symbols = []
            for symb in structure_copy.get_chemical_symbols():
                for group in map_occupations:  # pragma: no cover - because of the break
                    if symb in group:
                        new_symbols.append(group[0])
                        break
        else:
            new_symbols = len(structure) * ['H']
        structure_copy.set_chemical_symbols(new_symbols)
    structure_tuple = (
        structure_copy.get_cell(),
        structure_copy.get_scaled_positions(),
        structure_copy.numbers)
    dataset = spglib.get_symmetry_dataset(structure_tuple, symprec=symprec)
    n_unitcells = np.linalg.det(dataset.transformation_matrix)

    equivalent_atoms = list(dataset.equivalent_atoms)
    wyckoffs = {}
    for index in set(equivalent_atoms):
        multiplicity = list(dataset.equivalent_atoms).count(index) / n_unitcells
        multiplicity = int(round(multiplicity))
        wyckoff = '{}{}'.format(multiplicity, dataset.wyckoffs[index])
        if include_representative_atom_index:
            wyckoff += f'-{index}'
        wyckoffs[index] = wyckoff

    return [wyckoffs[equivalent_atoms[a.index]] for a in structure_copy]



from dataclasses import dataclass
from itertools import product

import numpy as np

NetworkWeights = dict[str, dict[str, np.ndarray]]
DescriptorWeights = dict[tuple[str, str], np.ndarray]
RestartParameters = dict[str, dict[str, dict[str, np.ndarray]]]


def _get_restart_contents(filename: str) -> tuple[list[float], list[float]]:
    """Parses a ``nep.restart`` file, and returns an unformatted list of the
    mean and standard deviation for all model parameters.
    Intended to be used by the py:meth:`~Model.read_restart` function.

    Parameters
    ----------
    filename
        input file name
    """
    mu = []     # Mean
    sigma = []  # Standard deviation
    with open(filename) as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            assert len(flds) != 0, f'Empty line number {k}'
            if len(flds) == 2:
                mu.append(float(flds[0]))
                sigma.append(float(flds[1]))
            else:
                raise IOError(f'Failed to parse line {k} from {filename}')
    return mu, sigma


def _get_model_type(first_row: list[str]) -> str:
    """Parses a the first row of a ``nep.txt`` file, and returns the
    type of NEP model. Available types are `potential`, `potential_with_charges`,
    `dipole`, and `polarizability`.

    Parameters
    ----------
    first_row
        First row of a NEP file, split by white space.
    """
    model_type = first_row[0]
    if 'charge' in model_type:
        return 'potential_with_charges'
    elif 'dipole' in model_type:
        return 'dipole'
    elif 'polarizability' in model_type:
        return 'polarizability'
    return 'potential'


def _get_nep_contents(filename: str) -> tuple[dict, list[float]]:
    """Parses a ``nep.txt`` file, and returns a dict describing the header
    and an unformatted list of all model parameters.
    Intended to be used by the :func:`read_model <calorine.nep.read_model>` function.

    Parameters
    ----------
    filename
        input file name
    """
    # parse file and split header and parameters
    header = []
    parameters = []
    nheader = 5  # 5 rows for NEP2, 6-7 rows for NEP3 onwards
    base_line = 3
    with open(filename) as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            assert len(flds) != 0, f'Empty line number {k}'
            if k == 0 and 'zbl' in flds[0]:
                base_line += 1
                nheader += 1
            if k == base_line and 'basis_size' in flds[0]:
                # Introduced in nep.txt after GPUMD v3.2
                nheader += 1
            if k < nheader:
                header.append(tuple(flds))
            elif len(flds) == 1:
                parameters.append(float(flds[0]))
            else:
                raise IOError(f'Failed to parse line {k} from {filename}')
    # compile data from the header into a dict
    data = {}
    for flds in header:
        if flds[0] in ['cutoff', 'zbl']:
            data[flds[0]] = tuple(map(float, flds[1:]))
        elif flds[0] in ['n_max', 'l_max', 'ANN', 'basis_size']:
            data[flds[0]] = tuple(map(int, flds[1:]))
        elif flds[0].startswith('nep'):
            version = flds[0].replace('nep', '').split('_')[0]
            version = int(version)
            data['version'] = version
            data['types'] = flds[2:]
            data['model_type'] = _get_model_type(flds)
        else:
            raise ValueError(f'Unknown field: {flds[0]}')
    return data, parameters


def _sort_descriptor_parameters(parameters: list[float],
                                types: list[str],
                                n_max_radial: int,
                                n_basis_radial: int,
                                n_max_angular: int,
                                n_basis_angular: int) -> tuple[DescriptorWeights,
                                                               DescriptorWeights]:
    """Reads a list of descriptors parameters and sorts them into two
    appropriately structured `dicts`, one for radial and one for angular descriptor weights.
    Intended to be used by the :func:`read_model <calorine.nep.read_model>` function.
    """
    # split up descriptor by chemical species and radial/angular
    n_types = len(types)
    n = len(parameters) / (n_types * n_types)
    assert n.is_integer(), 'number of descriptor groups must be an integer'
    n = int(n)

    m = (n_max_radial + 1) * (n_basis_radial + 1)
    descriptor_weights = parameters.reshape((n, n_types * n_types)).T
    descriptor_weights_radial = descriptor_weights[:, :m]
    descriptor_weights_angular = descriptor_weights[:, m:]

    # add descriptors to data dict
    radial_descriptor_weights = {}
    angular_descriptor_weights = {}
    m = -1
    for i, j in product(range(n_types), repeat=2):
        m += 1
        s1, s2 = types[i], types[j]
        radial_descriptor_weights[(s1, s2)] = descriptor_weights_radial[m, :].reshape(
            (n_max_radial + 1, n_basis_radial + 1)
        )
        angular_descriptor_weights[(s1, s2)] = descriptor_weights_angular[m, :].reshape(
            (n_max_angular + 1, n_basis_angular + 1)
        )
    return radial_descriptor_weights, angular_descriptor_weights


def _sort_ann_parameters(parameters: list[float],
                         ann_groupings: list[str],
                         n_neuron: int,
                         n_networks: int,
                         n_bias: int,
                         n_descriptor: int,
                         is_polarizability_model: bool,
                         is_model_with_charges: bool
                         ) -> NetworkWeights:
    """Reads a list of model parameters and sorts them into an appropriately structured `dict`.
    Intended to be used by the :func:`read_model <calorine.nep.read_model>` function.
    """
    n_ann_input_weights = (n_descriptor + 1) * n_neuron  # weights + bias
    n_ann_output_weights = 2*n_neuron if is_model_with_charges else n_neuron  # only weights
    n_ann_parameters = (
        n_ann_input_weights + n_ann_output_weights
    ) * n_networks + n_bias

    # Group ANN parameters
    pars = {}
    n1 = 0
    n_network_params = n_ann_input_weights + n_ann_output_weights  # except last bias(es)

    n_count = 2 if is_polarizability_model else 1
    n_outputs = 2 if is_model_with_charges else 1
    for count in range(n_count):
        # if polarizability model, all parameters including bias are repeated
        # need to offset n1 by +1 to handle bias
        n1 += count
        for s in ann_groupings:
            # Get the parameters for the ANN; in the case of NEP4, there is effectively
            # one network per atomic species.
            ann_parameters = parameters[n1 : n1 + n_network_params]
            ann_input_weights = ann_parameters[:n_ann_input_weights]
            w0 = np.zeros((n_neuron, n_descriptor))
            w0[...] = np.nan
            b0 = np.zeros((n_neuron, 1))
            b0[...] = np.nan
            for n in range(n_neuron):
                for nu in range(n_descriptor):
                    w0[n, nu] = ann_input_weights[n * n_descriptor + nu]
            b0[:, 0] = ann_input_weights[n_neuron * n_descriptor :]

            assert np.all(
                w0.shape == (n_neuron, n_descriptor)
            ), f'w0 has invalid shape for key {s}; please submit a bug report'
            assert np.all(
                b0.shape == (n_neuron, 1)
            ), f'b0 has invalid shape for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(w0)
            ), f'some weights in w0 are nan for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(b0)
            ), f'some weights in b0 are nan for key {s}; please submit a bug report'

            ann_output_weights = ann_parameters[
                n_ann_input_weights : n_ann_input_weights + n_ann_output_weights
            ]
            w1 = np.zeros((1, n_neuron * n_outputs))
            w1[0, :] = ann_output_weights[:]
            assert np.all(
                w1.shape == (1, n_neuron * n_outputs)
            ), f'w1 has invalid shape for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(w1)
            ), f'some weights in w1 are nan for key {s}; please submit a bug report'

            if count == 0 and n_outputs == 1:
                pars[s] = dict(w0=w0, b0=b0, w1=w1)
            elif count == 0 and n_outputs == 2:
                pars[s] = dict(w0=w0, b0=b0, w1=w1[0, :n_neuron], w1_charge=w1[0, n_neuron:])
            else:
                pars[s].update({'w0_polar': w0, 'b0_polar': b0, 'w1_polar': w1})
            # Jump to bias
            n1 += n_network_params
            if n_bias > 1 and not is_model_with_charges:
                # For NEP5 models we additionally have one bias term per species.
                # Currently NEP5 only exists for potential models, but we'll
                # keep it here in case it gets added down the line.
                bias_label = 'b1' if count == 0 else 'b1_polar'
                pars[s][bias_label] = parameters[n1]
                n1 += 1
        # For NEP3 and NEP4 we only have one bias.
        # For NEP4 with charges we have two biases.
        # For NEP5 we have one bias per species, and one global.
        if count == 0 and n_outputs == 1:
            pars['b1'] = parameters[n1]
        elif count == 0 and n_outputs == 2:
            pars['sqrt_epsilon_infinity'] = parameters[n1]
            pars['b1'] = parameters[n1+1]
        else:
            pars['b1_polar'] = parameters[n1]
    sum = 0
    for s in pars.keys():
        if s.startswith('b1') or s.startswith('sqrt'):
            sum += 1
        else:
            sum += np.sum([np.array(p).size for p in pars[s].values()])
    assert sum == n_ann_parameters * n_count, (
        'Inconsistent number of parameters accounted for; please submit a bug report\n'
        f'{sum} != {n_ann_parameters}'
    )
    return pars


@dataclass
class Model:
    r"""Objects of this class represent a NEP model in a form suitable for
    inspection and manipulation. Typically a :class:`Model` object is instantiated
    by calling the :func:`read_model <calorine.nep.read_model>` function.

    Attributes
    ----------
    version : int
        NEP version.
    model_type: str
        One of ``potential``, ``dipole`` or ``polarizability``.
    types : tuple[str, ...]
        Chemical species that this model represents.
    radial_cutoff : float | list[float]
        The radial cutoff parameter in Å.
        Is a list of radial cutoffs ordered after ``types`` in the case of typewise cutoffs.
    angular_cutoff : float | list[float]
        The angular cutoff parameter in Å.
        Is a list of angular cutoffs ordered after ``types`` in the case of typewise cutoffs.
    max_neighbors_radial : int
        Maximum number of neighbors in neighbor list for radial terms.
    max_neighbors_angular : int
        Maximum number of neighbors in neighbor list for angular terms.
    radial_typewise_cutoff_factor : float
        The radial cutoff factor if use_typewise_cutoff is used.
    angular_typewise_cutoff_factor : float
        The angular cutoff factor if use_typewise_cutoff is used.
    zbl : tuple[float, float]
        Inner and outer cutoff for transition to ZBL potential.
    zbl_typewise_cutoff_factor : float
        Typewise cutoff when use_typewise_cutoff_zbl is used.
    n_basis_radial : int
        Number of radial basis functions :math:`n_\mathrm{basis}^\mathrm{R}`.
    n_basis_angular : int
        Number of angular basis functions :math:`n_\mathrm{basis}^\mathrm{A}`.
    n_max_radial : int
        Maximum order of Chebyshev polymonials included in
        radial expansion :math:`n_\mathrm{max}^\mathrm{R}`.
    n_max_angular : int
        Maximum order of Chebyshev polymonials included in
        angular expansion :math:`n_\mathrm{max}^\mathrm{A}`.
    l_max_3b : int
        Maximum expansion order for three-body terms :math:`l_\mathrm{max}^\mathrm{3b}`.
    l_max_4b : int
        Maximum expansion order for four-body terms :math:`l_\mathrm{max}^\mathrm{4b}`.
    l_max_5b : int
        Maximum expansion order for five-body terms :math:`l_\mathrm{max}^\mathrm{5b}`.
    n_descriptor_radial : int
        Dimension of radial part of descriptor.
    n_descriptor_angular : int
        Dimension of angular part of descriptor.
    n_neuron : int
        Number of neurons in hidden layer.
    n_parameters : int
        Total number of parameters including scalers (which are not fit parameters).
    n_descriptor_parameters : int
        Number of parameters in descriptor.
    n_ann_parameters : int
        Number of neural network weights.
    ann_parameters : dict[tuple[str, dict[str, np.darray]]]
        Neural network weights.
    q_scaler : List[float]
        Scaling parameters.
    radial_descriptor_weights : dict[tuple[str, str], np.ndarray]
        Radial descriptor weights by combination of species; the array for each combination
        has dimensions of
        :math:`(n_\mathrm{max}^\mathrm{R}+1) \times (n_\mathrm{basis}^\mathrm{R}+1)`.
    angular_descriptor_weights : dict[tuple[str, str], np.ndarray]
        Angular descriptor weights by combination of species; the array for each combination
        has dimensions of
        :math:`(n_\mathrm{max}^\mathrm{A}+1) \times (n_\mathrm{basis}^\mathrm{A}+1)`.
    sqrt_epsilon_infinity : Optional[float]
        Square root of epsilon infinity $\epsilon_\infty$ (only for NEP models with charges).
    restart_parameters :  dict[str, dict[str, dict[str, np.ndarray]]]
        NEP restart parameters. A nested dictionary that contains the mean (mu) and standard
        deviation (sigma) for the ANN and descriptor parameters. Is set using the
        py:meth:`~Model.read_restart` method. Defaults to None.
    """

    version: int
    model_type: str
    types: tuple[str, ...]

    radial_cutoff: float | list[float]
    angular_cutoff: float | list[float]

    n_basis_radial: int
    n_basis_angular: int
    n_max_radial: int
    n_max_angular: int
    l_max_3b: int
    l_max_4b: int
    l_max_5b: int
    n_descriptor_radial: int
    n_descriptor_angular: int

    n_neuron: int
    n_parameters: int
    n_descriptor_parameters: int
    n_ann_parameters: int
    ann_parameters: NetworkWeights
    q_scaler: list[float]
    radial_descriptor_weights: DescriptorWeights
    angular_descriptor_weights: DescriptorWeights
    sqrt_epsilon_infinity: float = None
    restart_parameters: RestartParameters = None

    zbl: tuple[float, float] = None
    zbl_typewise_cutoff_factor: float = None
    max_neighbors_radial: int = None
    max_neighbors_angular: int = None
    radial_typewise_cutoff_factor: float = None
    angular_typewise_cutoff_factor: float = None

    _special_fields = [
        'ann_parameters',
        'q_scaler',
        'radial_descriptor_weights',
        'angular_descriptor_weights',
    ]

    def __str__(self) -> str:
        s = []
        for fld in self.__dataclass_fields__:
            if fld not in self._special_fields:
                s += [f'{fld:22} : {getattr(self, fld)}']
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<table border="1" class="dataframe"']
        s += [
            '<thead><tr><th style="text-align: left;">Field</th><th>Value</th></tr></thead>'
        ]
        s += ['<tbody>']
        for fld in self.__dataclass_fields__:
            if fld not in self._special_fields:
                s += [
                    f'<tr><td style="text-align: left;">{fld:22}</td>'
                    f'<td>{getattr(self, fld)}</td><tr>'
                ]
        for fld in self._special_fields:
            d = getattr(self, fld)
            # print('xxx', fld, d)
            if fld.endswith('descriptor_weights'):
                dim = list(d.values())[0].shape
            elif fld == 'ann_parameters' and self.version == 4:
                dim = (len(self.types), len(list(d.values())[0]))
            else:
                dim = len(d)
            s += [
                f'<tr><td style="text-align: left;">Dimension of {fld:22}</td><td>{dim}</td><tr>'
            ]
        s += ['</tbody>']
        s += ['</table>']
        return ''.join(s)

    def remove_species(self, species: list[str]):
        """Removes one or more species from the model.

        This method modifies the model in-place by removing all parameters
        associated with the specified chemical species. It prunes the species
        list, the Artificial Neural Network (ANN) parameters, and the
        descriptor weights. It also recalculates the total number of
        parameters in the model.

        Parameters
        ----------
        species
            A list of species names (str) to remove from the model.

        Raises
        ------
        ValueError
            If any of the provided species is not found in the model.
        """
        for s in species:
            if s not in self.types:
                raise ValueError(f'{s} is not a species supported by the NEP model')

        # --- Prune attributes based on species ---
        types_to_keep = [t for t in self.types if t not in species]
        self.types = tuple(types_to_keep)

        # Prune ANN parameters (for NEP4 and NEP5)
        if self.version in [4, 5]:
            self.ann_parameters = {
                key: value for key, value in self.ann_parameters.items()
                if key in types_to_keep or key.startswith('b1')
            }

        # Prune descriptor weights
        # key is here a tuple, (species1, species2)
        self.radial_descriptor_weights = {
            key: value for key, value in self.radial_descriptor_weights.items()
            if key[0] in types_to_keep and key[1] in types_to_keep
        }
        self.angular_descriptor_weights = {
            key: value for key, value in self.angular_descriptor_weights.items()
            if key[0] in types_to_keep and key[1] in types_to_keep
        }

        # Prune restart parameters if they have been loaded
        if self.restart_parameters is not None:
            for param_type in ['mu', 'sigma']:
                # Prune ANN restart parameters
                ann_key = f'ann_{param_type}'
                if self.version in [4, 5]:
                    self.restart_parameters[ann_key] = {
                        key: value for key, value in self.restart_parameters[ann_key].items()
                        if key in types_to_keep or key.startswith('b1')
                    }

                # Prune descriptor restart parameters
                for desc_type in ['radial', 'angular']:
                    key = f'{desc_type}_descriptor_{param_type}'
                    self.restart_parameters[key] = {
                        k: v for k, v in self.restart_parameters[key].items()
                        if k[0] in types_to_keep and k[1] in types_to_keep
                    }

        # --- Recalculate parameter counts ---
        n_types = len(self.types)
        n_descriptor = self.n_descriptor_radial + self.n_descriptor_angular

        # Recalculate descriptor parameter count
        self.n_descriptor_parameters = n_types**2 * (
            (self.n_max_radial + 1) * (self.n_basis_radial + 1)
            + (self.n_max_angular + 1) * (self.n_basis_angular + 1)
        )

        # Recalculate ANN parameter count
        if self.version == 3:
            n_networks = 1
            n_bias = 1
        elif self.version == 4:
            n_networks = n_types
            n_bias = 1
        else:  # NEP5
            n_networks = n_types
            n_bias = 1 + n_types

        n_ann_input_weights = (n_descriptor + 1) * self.n_neuron
        n_ann_output_weights = self.n_neuron
        self.n_ann_parameters = (
            n_ann_input_weights + n_ann_output_weights
        ) * n_networks + n_bias

        # Recalculate total parameter count
        self.n_parameters = (
            self.n_ann_parameters
            + self.n_descriptor_parameters
            + n_descriptor  # q_scaler parameters
        )
        if self.model_type == 'polarizability':
            self.n_parameters += self.n_ann_parameters

    def write(self, filename: str) -> None:
        """Write NEP model to file in `nep.txt` format."""
        with open(filename, 'w') as f:
            # header
            version_name = f'nep{self.version}'
            if self.zbl is not None:
                version_name += '_zbl'
            elif self.model_type != 'potential':
                version_name += f'_{self.model_type}'
            f.write(f'{version_name} {len(self.types)} {" ".join(self.types)}\n')
            if self.zbl is not None:
                f.write(f'zbl {" ".join(map(str, self.zbl))}\n')
            f.write('cutoff')
            if isinstance(self.radial_cutoff, float) and isinstance(self.angular_cutoff, float):
                f.write(f' {self.radial_cutoff} {self.angular_cutoff}')
            else:
                # Typewise cutoffs: one set of cutoffs per type
                for i in range(len(self.types)):
                    f.write(f' {self.radial_cutoff[i]} {self.angular_cutoff[i]}')
            f.write(f' {self.max_neighbors_radial} {self.max_neighbors_angular}')
            f.write('\n')
            f.write(f'n_max {self.n_max_radial} {self.n_max_angular}\n')
            f.write(f'basis_size {self.n_basis_radial} {self.n_basis_angular}\n')
            f.write(f'l_max {self.l_max_3b} {self.l_max_4b} {self.l_max_5b}\n')
            f.write(f'ANN {self.n_neuron} 0\n')

            # neural network weights
            keys = self.types if self.version in (4, 5) else ['all_species']
            suffixes = ['', '_polar'] if self.model_type == 'polarizability' else ['']
            for suffix in suffixes:
                for s in keys:
                    # Order: w0, b0, w1 (, b1 if NEP5)
                    # w0 indexed as: n*N_descriptor + nu
                    w0 = self.ann_parameters[s][f'w0{suffix}']
                    b0 = self.ann_parameters[s][f'b0{suffix}']
                    w1 = self.ann_parameters[s][f'w1{suffix}']
                    for n in range(self.n_neuron):
                        for nu in range(
                            self.n_descriptor_radial + self.n_descriptor_angular
                        ):
                            f.write(f'{w0[n, nu]:15.7e}\n')
                    for b in b0[:, 0]:
                        f.write(f'{b:15.7e}\n')
                    for v in w1[0, :]:
                        f.write(f'{v:15.7e}\n')
                    if self.version == 5:
                        b1 = self.ann_parameters[s][f'b1{suffix}']
                        f.write(f'{b1:15.7e}\n')
                b1 = self.ann_parameters[f'b1{suffix}']
                f.write(f'{b1:15.7e}\n')

            # descriptor weights
            mat = []
            for s1 in self.types:
                for s2 in self.types:
                    mat = np.hstack(
                        [mat, self.radial_descriptor_weights[(s1, s2)].flatten()]
                    )
                    mat = np.hstack(
                        [mat, self.angular_descriptor_weights[(s1, s2)].flatten()]
                    )
            n_types = len(self.types)
            n = int(len(mat) / (n_types * n_types))
            mat = mat.reshape((n_types * n_types, n)).T
            for v in mat.flatten():
                f.write(f'{v:15.7e}\n')

            # scaler
            for v in self.q_scaler:
                f.write(f'{v:15.7e}\n')

    def read_restart(self, filename: str):
        """Parses a file in `nep.restart` format and saves the
        content in the form of mean and standard deviation for each
        parameter in the corresponding NEP model.

        Parameters
        ----------
        filename
            Input file name.
        """
        mu, sigma = _get_restart_contents(filename)
        restart_parameters = np.array([mu, sigma]).T

        is_polarizability_model = self.model_type == 'polarizability'
        is_charged_model = self.model_type == 'potential_with_charges'

        n1 = self.n_ann_parameters
        n1 *= 2 if is_polarizability_model else 1
        n2 = n1 + self.n_descriptor_parameters
        ann_parameters = restart_parameters[:n1]
        descriptor_parameters = np.array(restart_parameters[n1:n2])

        if self.version == 3:
            n_networks = 1
            n_bias = 1
        elif self.version == 4:
            # one hidden layer per atomic species
            n_networks = len(self.types)
            n_bias = 1
        else:
            raise ValueError(f'Cannot load nep.restart for NEP model version {self.version}')

        ann_groups = [s for s in self.ann_parameters.keys() if not s.startswith('b1')]
        n_bias = len([s for s in self.ann_parameters.keys() if s.startswith('b1')])
        n_descriptor = self.n_descriptor_radial + self.n_descriptor_angular
        restart = {}

        for i, content_type in enumerate(['mu', 'sigma']):
            ann = _sort_ann_parameters(ann_parameters[:, i],
                                       ann_groups,
                                       self.n_neuron,
                                       n_networks,
                                       n_bias,
                                       n_descriptor,
                                       is_polarizability_model,
                                       is_charged_model)
            radial, angular = _sort_descriptor_parameters(descriptor_parameters[:, i],
                                                          self.types,
                                                          self.n_max_radial,
                                                          self.n_basis_radial,
                                                          self.n_max_angular,
                                                          self.n_basis_angular)

            restart[f'ann_{content_type}'] = ann
            restart[f'radial_descriptor_{content_type}'] = radial
            restart[f'angular_descriptor_{content_type}'] = angular
        self.restart_parameters = restart

    def write_restart(self, filename: str):
        """Write NEP restart parameters to file in `nep.restart` format."""
        keys = self.types if self.version in (4, 5) else ['all_species']
        suffixes = ['', '_polar'] if self.model_type == 'polarizability' else ['']
        columns = []
        for i, parameter in enumerate(['mu', 'sigma']):
            # neural network weights
            ann_parameters = self.restart_parameters[f'ann_{parameter}']
            column = []
            for suffix in suffixes:
                for s in keys:
                    # Order: w0, b0, w1 (, b1 if NEP5)
                    # w0 indexed as: n*N_descriptor + nu
                    w0 = ann_parameters[s][f'w0{suffix}']
                    b0 = ann_parameters[s][f'b0{suffix}']
                    w1 = ann_parameters[s][f'w1{suffix}']
                    for n in range(self.n_neuron):
                        for nu in range(
                            self.n_descriptor_radial + self.n_descriptor_angular
                        ):
                            column.append(f'{w0[n, nu]:15.7e}')
                    for b in b0[:, 0]:
                        column.append(f'{b:15.7e}')
                    for v in w1[0, :]:
                        column.append(f'{v:15.7e}')
                b1 = ann_parameters[f'b1{suffix}']
                column.append(f'{b1:15.7e}')
            columns.append(column)

            # descriptor weights
            radial_descriptor_parameters = self.restart_parameters[f'radial_descriptor_{parameter}']
            angular_descriptor_parameters = self.restart_parameters[
                    f'angular_descriptor_{parameter}']
            mat = []
            for s1 in self.types:
                for s2 in self.types:
                    mat = np.hstack(
                        [mat, radial_descriptor_parameters[(s1, s2)].flatten()]
                    )
                    mat = np.hstack(
                        [mat, angular_descriptor_parameters[(s1, s2)].flatten()]
                    )
            n_types = len(self.types)
            n = int(len(mat) / (n_types * n_types))
            mat = mat.reshape((n_types * n_types, n)).T
            for v in mat.flatten():
                column.append(f'{v:15.7e}')

        # Join the mean and standard deviation columns
        assert len(columns[0]) == len(columns[1]), 'Length of means must match standard deviation'
        joined = [f'{s1} {s2}\n' for s1, s2 in zip(*columns)]
        with open(filename, 'w') as f:
            f.writelines(joined)


def read_model(filename: str) -> Model:
    """Parses a file in ``nep.txt`` format and returns the
    content in the form of a :class:`Model <calorine.nep.model.Model>`
    object.

    Parameters
    ----------
    filename
        Input file name.
    """
    data, parameters = _get_nep_contents(filename)

    # sanity checks
    for fld in ['cutoff', 'basis_size', 'n_max', 'l_max', 'ANN']:
        assert fld in data, f'Invalid model file; {fld} line is missing'
    assert data['version'] in [
        3,
        4,
        5,
    ], 'Invalid model file; only NEP versions 3, 4 and 5 are currently supported'

    # split up cutoff tuple
    N_types = len(data['types'])
    # Either global cutoffs + max neighbirs, or typewise cutoffs + max_neighbors
    assert len(data['cutoff']) in [4, 2*N_types+2]
    data['max_neighbors_radial'] = int(data['cutoff'][-2])
    data['max_neighbors_angular'] = int(data['cutoff'][-1])
    if len(data['cutoff']) == 2*N_types+2:
        # Typewise cutoffs: radial are even, angular are odd
        data['radial_cutoff'] = [data['cutoff'][i*2] for i in range(N_types)]
        data['angular_cutoff'] = [data['cutoff'][i*2+1] for i in range(N_types)]
    else:
        data['radial_cutoff'] = data['cutoff'][0]
        data['angular_cutoff'] = data['cutoff'][1]
    del data['cutoff']

    # split up basis_size tuple
    assert len(data['basis_size']) == 2
    data['n_basis_radial'] = data['basis_size'][0]
    data['n_basis_angular'] = data['basis_size'][1]
    del data['basis_size']

    # split up n_max tuple
    assert len(data['n_max']) == 2
    data['n_max_radial'] = data['n_max'][0]
    data['n_max_angular'] = data['n_max'][1]
    del data['n_max']

    # split up nl_max tuple
    len_l = len(data['l_max'])
    assert len_l in [1, 2, 3]
    data['l_max_3b'] = data['l_max'][0]
    data['l_max_4b'] = data['l_max'][1] if len_l > 1 else 0
    data['l_max_5b'] = data['l_max'][2] if len_l > 2 else 0
    del data['l_max']

    # compute dimensions of descriptor components
    data['n_descriptor_radial'] = data['n_max_radial'] + 1
    l_max_enh = data['l_max_3b'] + (data['l_max_4b'] > 0) + (data['l_max_5b'] > 0)
    data['n_descriptor_angular'] = (data['n_max_angular'] + 1) * l_max_enh
    n_descriptor = data['n_descriptor_radial'] + data['n_descriptor_angular']

    is_charged_model = data['model_type'] == 'potential_with_charges'
    # compute number of parameters
    data['n_neuron'] = data['ANN'][0]
    del data['ANN']
    n_types = len(data['types'])
    if data['version'] == 3:
        n = 1
        n_bias = 1
    elif data['version'] == 4 and is_charged_model:
        # one hidden layer per atomic species, but two output nodes
        n = n_types
        n_bias = 2
    elif data['version'] == 4:
        # one hidden layer per atomic species
        n = n_types
        n_bias = 1
    else:  # NEP5
        # like nep4, but additionally has an
        # individual bias term in the output
        # layer for each species.
        n = n_types
        n_bias = 1 + n_types  # one global bias + one per species

    n_ann_input_weights = (n_descriptor + 1) * data['n_neuron']  # weights + bias
    n_ann_output_weights = 2*data['n_neuron'] if is_charged_model else data['n_neuron']  # weights
    n_ann_parameters = (
        n_ann_input_weights + n_ann_output_weights
    ) * n + n_bias

    n_descriptor_weights = n_types**2 * (
        (data['n_max_radial'] + 1) * (data['n_basis_radial'] + 1)
        + (data['n_max_angular'] + 1) * (data['n_basis_angular'] + 1)
    )
    data['n_parameters'] = n_ann_parameters + n_descriptor_weights + n_descriptor
    is_polarizability_model = data['model_type'] == 'polarizability'
    if data['n_parameters'] + n_ann_parameters == len(parameters):
        data['n_parameters'] += n_ann_parameters
        assert is_polarizability_model, (
            'Model is not labelled as a polarizability model, but the number of '
            'parameters matches a polarizability model.\n'
            'If this is a polarizability model trained with GPUMD <=v3.8, please '
            'modify the header in the nep.txt file to enable parsing '
            f'`nep{data["version"]}_polarizability`.\n'
        )
    assert data['n_parameters'] == len(parameters), (
        'Parsing of parameters inconsistent; please submit a bug report\n'
        f'{data["n_parameters"]} != {len(parameters)}'
    )
    data['n_ann_parameters'] = n_ann_parameters

    # split up parameters into the ANN weights, descriptor weights, and scaling parameters
    n1 = n_ann_parameters
    n1 *= 2 if is_polarizability_model else 1
    n2 = n1 + n_descriptor_weights
    data['ann_parameters'] = parameters[:n1]
    descriptor_weights = np.array(parameters[n1:n2])
    data['q_scaler'] = parameters[n2:]

    # add ann parameters to data dict
    ann_groups = data['types'] if data['version'] in (4, 5) else ['all_species']
    sorted_ann_parameters = _sort_ann_parameters(data['ann_parameters'],
                                                 ann_groups,
                                                 data['n_neuron'],
                                                 n,
                                                 n_bias,
                                                 n_descriptor,
                                                 is_polarizability_model,
                                                 is_charged_model)

    data['ann_parameters'] = sorted_ann_parameters
    if 'sqrt_epsilon_infinity' in sorted_ann_parameters.keys():
        data['sqrt_epsilon_infinity'] = sorted_ann_parameters['sqrt_epsilon_infinity']
        sorted_ann_parameters.pop('sqrt_epsilon_infinity')
        data['ann_parameters'] = sorted_ann_parameters

    # add descriptors to data dict
    data['n_descriptor_parameters'] = len(descriptor_weights)
    radial, angular = _sort_descriptor_parameters(descriptor_weights,
                                                  data['types'],
                                                  data['n_max_radial'],
                                                  data['n_basis_radial'],
                                                  data['n_max_angular'],
                                                  data['n_basis_angular'])
    data['radial_descriptor_weights'] = radial
    data['angular_descriptor_weights'] = angular

    return Model(**data)



from warnings import warn
import numpy as np

try:
    import torch
    torch_available = True
except ImportError:  # pragma: no cover
    torch_available = False
    # If pytorch is not available, we still need a torch object
    # to satisfy type hints etc. Otherwise the file will not run.

    class _DummyTensor:
        """Fallback Tensor placeholder."""
        pass

    class _DummyDevice:
        """Fallback device placeholder."""
        def __init__(self, *args, **kwargs):
            pass

    class _DummyDType:
        """Fallback dtype placeholder."""
        pass

    class _DummyTorchModule:
        Tensor = _DummyTensor
        device = _DummyDevice
        dtype = _DummyDType

        # Provide common dtype names you might reference
        float32 = _DummyDType()
        float64 = _DummyDType()

    torch = _DummyTorchModule()

try:
    from tqdm.auto import tqdm
except ImportError:  # pragma: no cover
    def tqdm(x):
        return x


def get_entropy(
    descriptors: np.ndarray | torch.Tensor,
    width: float,
    use_tqdm: bool = True,
    device: str | torch.device = 'cuda',
    block: int = 1024,
    dtype: torch.dtype = torch.float32,
    eps: float = 1e-12,
) -> tuple[float, np.ndarray]:
    r"""
    Computes an estimate for the information entropy :math:`H(\mathbf{X})` of a
    set of descriptors :math:`\mathbf{X}`. The estimate is described in
    [Nat. Comm. **16**, 4014 (2025)](https://doi.org/10.1038/s41467-025-59232-0)
    and given by

    .. math::

        \mathcal{H}(\{\mathbf{X}\}) = -\frac{1}{n} \sum_{i=1}^{n} p_i

    where

    .. math::

        p_i
        = \log \left[
                \frac{1}{n} \sum_{j=1}^{n}
                K_h(\mathbf{X}_i, \mathbf{X}_j)
            \right]

    with a Gaussian kernel

    .. math::

        K_h(\mathbf{X}_i, \mathbf{X}_j)
        = \exp\!\left(
            -\frac{\lVert \mathbf{X}_i - \mathbf{X}_j \rVert^2}{2h^2}.
        \right)

    The calculation is done via torch if the library has been installed,
    and numpy otherwise. When using torch the calculation is run via CUDA.
    The latter behavior can be controlled using the :attr:`device` argument.

    Parameters
    ----------
    descriptors
        The set of descriptors :math:`\mathbf{X}` for which to evaluate to
        the entropy. Typically each row corresponds to one atom and the
        columns correspond to the different descriptor components.
    width
        Width :math:`h` of the Gaussian kernel.
    use_tqdm
        Use `tqdm <https://tqdm.github.io/>`_ to show a progress bar.
        Note that this requires tqdm to be installed.
    block
        In order to limit the memory needs, the kernel density estimate
        matrix is handled in blocks. This parameter controls the size of
        each block. Smaller numbers imply a smaller memory footprint.
    eps
        Smallest (absolute) permissible value.
    device
        Device to use for calculation. The documentation of
        [`torch.device`](https://docs.pytorch.org/docs/stable/tensor_attributes.html#torch.device)
        provides more information.
        Only used when pytorch is available.
    dtype
        Floating point precision used for the computation. The documentation of
        [`torch.dtype`](https://docs.pytorch.org/docs/stable/tensor_attributes.html)
        provides an overview.
        Only used when pytorch is available.

    Returns
    -------
    A tuple comprising the total entropy :math:`H(\mathbf{X})` and the entropy contributions
    :math:`p_i` from each row in the input descriptor matrix.
    """
    if torch_available:
        res = _get_entropy_torch(descriptors, width, use_tqdm, block, eps, device, dtype)
        return res
    else:  # pragma: no cover
        warn('Using the numpy implementation.'
             ' Install torch in order to use GPUs and achieve a considerable speed-up.')
        res = _get_entropy_numpy(descriptors, width, use_tqdm, block, eps)
        return res


def _get_entropy_numpy(
    descriptors: np.ndarray,
    width: float,
    use_tqdm: bool,
    block: int,
    eps: float,
) -> float:
    """Compute the informational entropy using numpy.
    See get_entropy for documentation.
    """
    X = np.asarray(descriptors, dtype=np.float64)
    N, d = X.shape
    s = np.sum(X * X, axis=1)                         # (N,)
    inv_two_sigma2 = 1.0 / (2.0 * width * width)

    row_sums = np.zeros(N, dtype=np.float64)
    for i0 in tqdm(range(0, N, block), leave=False):
        i1 = min(i0 + block, N)
        # D2[i0:i1, :] = s[i0:i1,None] + s[None,:] - 2*X[i0:i1]@X.T
        G = X[i0:i1] @ X.T                            # (B,N)
        D2_blk = (s[i0:i1, None] + s[None, :] - 2.0 * G)
        row_sums[i0:i1] = np.sum(np.exp(-D2_blk * inv_two_sigma2), axis=1)

    subvals = -np.log(np.clip(row_sums / N, min=eps))
    subvals /= N
    entropy = np.sum(subvals)
    return -float(entropy), subvals


def _get_entropy_torch(
    descriptors: np.ndarray | torch.Tensor,
    width: float,
    use_tqdm: bool,
    block: int,
    eps: float,
    device: str | torch.device = 'cuda',
    dtype: torch.dtype = torch.float32,
) -> tuple[float, np.ndarray]:
    """Compute the informational entropy using torch.
    See get_entropy for documentation.
    """
    with torch.no_grad():
        # Move data to device
        if not torch.cuda.is_available() and str(device) == 'cuda':  # pragma: no cover
            device = 'cpu'
        X = torch.as_tensor(descriptors, dtype=dtype, device=device)
        N, d = X.shape
        inv_two_sigma2 = 1.0 / (2.0 * width * width)

        # Precompute norms once
        s = (X * X).sum(dim=1)  # (N,)
        row_sums = torch.zeros(N, dtype=dtype, device=device)

        # Block over rows i; each block computes K[i0:i1, :].sum(-1)
        # D2 = s[i] + s[j] - 2 * X[i] @ X[j]^T  (formed in blocks to save memory)
        XT = X.T  # reuse in matmuls
        for i0 in tqdm(range(0, N, block), leave=False):
            i1 = min(i0 + block, N)
            try:
                G = X[i0:i1] @ XT                       # (B, N)
                D2_blk = s[i0:i1, None] + s[None, :] - 2.0 * G
            except torch.cuda.OutOfMemoryError:  # pragma: no cover
                torch.cuda.empty_cache()
                raise ValueError(
                    'Tried to allocate too much GPU memory.'
                    f' Try to reduce the value of block, e.g., to {block//2}.')
            row_sums[i0:i1] = torch.exp(-D2_blk * inv_two_sigma2).sum(dim=1)

        subvals = -torch.log(torch.clamp(row_sums / N, min=eps))
        subvals /= N
        entropy = torch.sum(subvals)
    return -float(entropy.detach().cpu().item()), subvals.detach().cpu().numpy()
